/**
 * Application  LTCPMON - SIFTR to L3DGEWorld Interface Utility
 *
 * Revision     1.0a (9th August 2010)
 * Programmer   Michael Allwright
 *
 * Purpose      LTCPMON is an application designed to parse SIFTR log files and configurably project parameters
 *              in these log files onto visual metrics in L3DGEWorld.
 *
 * TODO / BUG   Use proper unix time functions - ctime::time(null) only returns seconds, milliseconds required...
**/

#include <iostream>
#include <map>

#include <cstdlib> // for atoi and atof etc
#include <ctime>

#include <getopt.h>
#include <unistd.h>

#include "siftrlogparser.h"
#include "mapfileparser.h"
#include "configuration.h"
#include "exceptions.h"
#include "tcpflow.h"
#include "tcpmapping.h"
#include "L3DGECommunicator.h"

/**
 * Method Name  void usage()
 *
 * Description  Prints the usage message to std::cerr
**/
void usage()
{
    std::cerr << "usage: ltcpmon [options] <map_file> <entity_file> <siftr_logfile>" << std::endl << std::endl
          << "options:" << std::endl
          << "\t[-s <time>] - visualisation start time offset since beginning of the SIFTR log (in seconds)" << std::endl
          << "\t[-d <duration>] - duration of visualisation (in seconds)" << std::endl
          << "\t[-u <period>] - period at which updates are sent to L3DGEWorld (in seconds)" << std::endl
          << "\t[-r <ratio>] - real:visualisation time ratio (visualisation speed)" << std::endl
          << "\t[-a <ip address>] - IP address of the L3DGEWorld Server (127.0.0.1)" << std::endl
          << "\t[-p <port>] - Communication port of the L3DGEWorld Server (27960)" << std::endl
          << "\t[-v] - Enable verbose mode" << std::endl << std::endl
          << "files:" << std::endl
          << "\t<map_file> - contains mappings of TCP parameters to visualisation metrics" << std::endl
          << "\t<entity_file> - contains a list of entities available in the server map" << std::endl
          << "\t<siftr_logfile> - contains the SIFTR data to be visualised" << std::endl << std::endl;
}

/**
 * Method Name  void printBannerMessage(const std::string & theMessage)
 *
 * Description  Prints the string in 'theMessage' to std::cerr in a box
**/
void printBannerMessage(const std::string & theMessage)
{
    std::cerr << "-------------------------------------------" << std::endl
              << theMessage << std::endl
              << "-------------------------------------------" << std::endl << std::endl;
}

/**
 * Method Name  void printInitInformation(const SiftrLogParser & siftrLogfile, const MapFileParser & mappingsFile)
 *
 * Description  Prints the results of initialisation to std::cerr. Includes information about the SIFTR logfile,
 *              the mappings file and various runtime parameters based on defaults or overridden values.
**/
void printInitInformation(const SiftrLogParser & siftrLogfile, const MapFileParser & mappingsFile)
{
    time_t enableTime = (time_t)siftrLogfile.getEnableTime();

    printBannerMessage("Daemon Initialisation Report and Statistics");
    std::cerr << std::endl
          << "L3DGEWorld Server" << std::endl
          << "\tIP: " << Configuration::serverIP << std::endl
          << "\tPort: " << Configuration::serverPort << std::endl
          << "Visualisation Playback Parameters" << std::endl
          << "\tStart time: " << Configuration::visualStartTime << " seconds" << std::endl
          << "\tRun time: ";
    if(Configuration::durationLimitFlag == false)
    {
    std::cerr << "All available data in logfile" << std::endl;
    }
    else
    {
    std::cerr << Configuration::visualDuration << " seconds" << std::endl;
    }
    std::cerr << "\tUpdate period: " << Configuration::updatePeriod << " seconds"  << std::endl
          << "\tPlayback Speed Ratio: " << Configuration::playbackRatio << std::endl
          << "Mappings File: " << mappingsFile.getFilename()  << std::endl
          << mappingsFile.generateReport() << std::endl
          << "SIFTR Logfile: " << Configuration::siftrLogFilename << std::endl
          << "\tSIFTR version: " << siftrLogfile.getSiftrVersion() << std::endl
          << "\tLogfile enable time: " << ctime(&enableTime)
          << "\tKernel Tick Rate: " << siftrLogfile.getKernelTickRate() << "Hz" << std::endl
          << "\tTCP RTT Scale: " << siftrLogfile.getTcpRttScale() << std::endl << std::endl;
}

/**
 * Method Name  void visualisationLoop()
 *
 * Description  Loads files and runs the main application loop as described:
 *                  1. Load files
 *                  2. Open connection (L3DGECommunicator)
 *                  3. Print some information
 *                  4. Parse data
 *                  5. Send updates to L3DGEWorld (in turn causes
 *                     stats to be calculated and scaled)
 *                  6. Check for more data and increment playback time variable
 *                  7. If more data goto 4, else exit.
**/
void visualisationLoop()
{
    std::map<std::string, TcpFlow> TcpFlowMap;
    const double timestep = (Configuration::updatePeriod * Configuration::playbackRatio);

    double playbackTimeInSeconds = Configuration::visualStartTime + timestep;

    /* Load the SIFTR log file */
    SiftrLogParser siftrLogfile(Configuration::siftrLogFilename);

    /* Load the Mappings file */
    MapFileParser mappingsFile(Configuration::mappingsFilename);
    std::vector<TcpMapping> mappingsList = mappingsFile.getMappings();

    /* Open the connection */
    L3DGECommunicator connection(Configuration::serverIP, Configuration::serverPort, Configuration::entityFilename);

    // Report Loading Statistics
    printInitInformation(siftrLogfile, mappingsFile);

    printBannerMessage("Starting Visualisation Replay Session");

    try
    {
        TcpParams toAdd;
        time_t lastUpdate = time(NULL); // initialise the last update to the current time.
        while(true)
        {
            std::cerr << std::endl << ":::: Current playback time = " << playbackTimeInSeconds << " seconds ::::" << std::endl;

            std::cerr << "Parsing visualisation data from " << Configuration::siftrLogFilename << std::endl;
            while(siftrLogfile.moreDataAvailable() == true)
            {
                toAdd = siftrLogfile.getNextFlowParameters();
                if(TcpFlowMap.find(toAdd.tcpFlowID) == TcpFlowMap.end()) // check whether flow is to be appended or created
                {
                    TcpFlowMap[toAdd.tcpFlowID] = TcpFlow(toAdd);
                }
                else
                {
                    TcpFlowMap[toAdd.tcpFlowID].appendStatsToFlow(toAdd);
                }
                if(toAdd.packetTimeStamp > (playbackTimeInSeconds + siftrLogfile.getEnableTime())) break; // parse up until playback time.
                else continue;
            }

            std::cerr << "Waiting for update time to expire..." << std::endl;
            while(time(NULL) < Configuration::updatePeriod + lastUpdate) usleep(200);

            std::cerr << "Updating L3DGEWorld Entities via L3DGECOMMS" << std::endl;
            connection.updateEntities(siftrLogfile.getEnableTime() + playbackTimeInSeconds - timestep,
                          TcpFlowMap, mappingsList);

            // mark the time when the last update occured
            time(&lastUpdate);

            playbackTimeInSeconds += timestep;

            // Visualisation update process termination conditions
            if(Configuration::durationLimitFlag == true)
            {
                if(playbackTimeInSeconds > (Configuration::visualDuration + Configuration::visualStartTime)) break;
            }
            if(siftrLogfile.moreDataAvailable() == false) break;

        }
    }
    catch(noMoreDataException e)
    {
        std::cerr.flush();
        std::cerr << e.what() << std::endl;
    }
}

/**
 * Method Name  int main(int argc, char * args[])
 *
 * Description  Parses the command line and configures options, passes control
 *              to the visualisationLoop method before exiting
**/
int main(int argc, char * args[])
{
    // Print banner
    std::cerr << std::endl << "LTCPMON - SIFTR to L3DGEWorld Interface Utility" << std::endl;

    // No arguments given, print usage.
    if(argc == 1)
    {
        usage();
        exit(0);
    }
    // Attempt initialisation (parsing of command line and loading of the provided siftr_logfile
    try
    {
        char option;
        opterr = 0; // Disable getopt.h from printing to stderr

        if(argc < 4) throw badSyntaxException("Initialisation Failed: The command line syntax was invalid");

        while ((option = getopt(argc, args, "s:d:u:r:a::p:v")) != -1)
        {
            switch(option)
            {
            case 's': Configuration::visualStartTime = atof(optarg);
                  break;
            case 'd': Configuration::visualDuration = atof(optarg);
                  Configuration::durationLimitFlag = true;
                  break;
            case 'u': Configuration::updatePeriod = atof(optarg);
                  break;
            case 'r': Configuration::playbackRatio = atof(optarg);
                  break;
            case 'a': Configuration::serverIP = std::string(optarg);
                  break;
            case 'p': Configuration::serverPort = atoi(optarg);
                  break;
            case 'v': Configuration::verboseMode = true;
                  break;
            default:  throw badSyntaxException("Initialisation Failed: The command line syntax was invalid");
                  break;
            }
        }
        /* Set log file from the last argument given */
        Configuration::siftrLogFilename = std::string(args[argc - 1]);

        /* Set entities file from the second last argument given */
        Configuration::entityFilename = std::string(args[argc - 2]);

        Configuration::mappingsFilename = std::string(args[argc - 3]);

        /* Start the visualisation control loop */
        visualisationLoop();
    }
    catch(siftrDaemonRuntimeException e)
    {
        std::cerr << std::endl << e.what() << std::endl << std::endl;
        if(e.printUsageFlagSet()) usage();
        if(e.exitRequiredFlagSet()) exit(1);
    }
    return 0;
}
