/**
 * C++ Class    Entity File Parser
 *
 * Revision     1.0a (9th August 2010)
 * Programmer   Michael Allwright
 *
 * Purpose      Entity File Parser is designed to parse a set of entities out of a given file and manage associations
 *              between the entities and the TCP flow.
 *
 *              The class contains several overloaded methods which allow you to set/get an association and to remove
 *              an association based on the entities UID or tcpflow ID (used in the flow map)...
 *
 * NOTE | BUG   If attempting to fix or improve the removal logic, check the reset association method logic here
 *              possible source for bugs...
**/

#include <fstream>
#include <sstream>
#include <iostream>
#include <cstring>

#include "entityfileparser.h"
#include "entitymapping.h"
#include "exceptions.h"

/**
 * Method Name  EntityFileParser::EntityFileParser(const std::string & entityFilename)
 *
 * Parameters   const std::string & entityFilename - A C++ string containing the entities file to be parsedLines
 *                                                   this file is typically known as l3dgehosts.conf
 *
 * Description  Constructor - sets up class by reseting counters to zero, clearing vectors, and initialising the
 *              object with a list of entities as parsed from the file 'entityFilename'
**/
EntityFileParser::EntityFileParser(const std::string & entityFilename) : entityFilename(entityFilename)
{
    (numberOfLines = 0, skippedLines = 0, commentLines = 0, parsedLines = 0);
    (numberOfEntities = 0, numberOfAssignedEntities = 0);
    entityFile.open(entityFilename.c_str());
    mappingsList.clear();
    entityFile.exceptions(std::ifstream::badbit); // enable internal exception throwing
    parse();
}

/**
 * Method Name  unsigned int EntityFileParser::setEntityAssociation(std::string tcpFlowID)
 *
 * Parameters   const std::string & tcpFlowID - A C++ string containing the TCP flow to be assigned to any entity
 *
 * Description  If already assigned, the method will return the existing UID for that flow. otherwise if there is
 *              room, this method will return a new UID for the flow. Alternatively an exception will be thrown if
 *              if there is no more UID to be assigned.
**/
unsigned int EntityFileParser::setEntityAssociation(const std::string & tcpFlowID)
{
    std::vector<EntityMapping>::iterator it;

    for(it = mappingsList.begin(); it != mappingsList.end(); it++) if((it->flowID.compare(tcpFlowID)) == 0) break;

    if(it != mappingsList.end()) // found match
    {
        return it->entityUID;
    }
    else for(it = mappingsList.begin(); it != mappingsList.end(); it++) if((it->flowID.compare("")) == 0) break;

    if(it != mappingsList.end()) // room to add flow
    {
        it->flowID.assign(tcpFlowID);
        numberOfAssignedEntities++;
        return it->entityUID;
    }
    else // no more room
        throw siftrDaemonRuntimeException("The number of TCP flows exceeds the number of entities",true,false);
}

/**
 * Method Name  bool EntityFileParser::removeEntityAssociation(unsigned int entityUID)
 *
 * Parameters   unsigned int entityUID - UID of the entity to be unassociated
 *
 * Description  Removes a flow from the entity association map based on the UID
 *              returns true if entity was found and removed. false if there was no record...
**/
bool EntityFileParser::removeEntityAssociation(unsigned int entityUID)
{
    std::vector<EntityMapping>::iterator it;

    for(it = mappingsList.begin(); it != mappingsList.end(); it++) if((it->entityUID) == entityUID) break;
    if(it != mappingsList.end())
    {
        it->flowID.assign("");
        return true;
    }
    else return false;
}

/**
 * Method Name  bool EntityFileParser::removeEntityAssociation(std::string tcpFlowID)
 *
 * Parameters   const std::string & tcpFlowID - flow ID of the tcp flow to be unassociated
 *
 * Description  Removes a flow from the entity association map based on the flow ID
 *              returns true if entity was found and removed. false if there was no record...
**/
bool EntityFileParser::removeEntityAssociation(const std::string & tcpFlowID)
{
    std::vector<EntityMapping>::iterator it;

    for(it = mappingsList.begin(); it != mappingsList.end(); it++) if((it->flowID.compare(tcpFlowID)) == 0) break;
    if(it != mappingsList.end())
    {
        it->flowID.assign("");
        return true;
    }
    else return false;
}

/**
 * Method Name  void EntityFileParser::parse()
 *
 * Description  Populates the class with available entities from the file 'entityFilename'
 *              (see constructor).
**/
void EntityFileParser::parse()
{
    unsigned int scanfArgsRead = 0;

    // default parameters
    unsigned int entityUID = 0;
    char entityName[41];
    std::string lineToParse;

    // reinitialise counters
    (numberOfLines = 0, skippedLines = 0, commentLines = 0, parsedLines = 0);
    (numberOfEntities = 0, numberOfAssignedEntities = 0);

    // Clear vector containing the mappings
    mappingsList.clear();

    if(entityFile.is_open())
    {
    try
    {
        while(entityFile.eof() == false)
        {
        getline(entityFile, lineToParse); // get the next line
        numberOfLines++;

        if(lineToParse[0] == '#') // skip comments
        {
            commentLines++;
            continue;
        }
        scanfArgsRead = sscanf(lineToParse.c_str(),"%40s %d %*d", entityName, &entityUID); // Parse it

        if(scanfArgsRead != 2) // check whether the line parsed sucessfully
        {
            skippedLines++;
            continue;
        }

        mappingsList.push_back(EntityMapping(entityName,entityUID,""));
        (parsedLines++, numberOfEntities++);
        }
    }
    catch(std::ios_base::failure e)
    {
        std::ostringstream errorMsg;
        errorMsg << "The entity file \"" << entityFilename << "\" is no longer accessible";
        throw fileAccessDeniedException(errorMsg.str());
    }
    }
    else
    {
    std::ostringstream errorMsg;
    errorMsg << "The entity file \"" << entityFilename << "\" could not be found";
    throw fileNotFoundException(errorMsg.str());
    }
}

