/**
 * C++ Class    Map File Parser
 *
 * Revision     1.0a (9th August 2010)
 * Programmer   Michael Allwright
 *
 * Purpose      Map File Parser is designed to parse the mappings between a network characteristic and a visualisation
 *              parameter. When parsing the class ingnores blank lines and lines beginning with '#' (commments),
 *              looking for lines matching the format:
 *              
 *              networkCharacteristic:statisticsMode:visualisationParameter
 *
 *              The statistic mode determines how the data will be processed before it is scaled. i.e. from of the data set
 *              find the minimum, maximum, average etc...
 *
**/

#include <fstream>
#include <sstream>

#include <cstring>
#include <iostream>

#include "configuration.h"
#include "mapfileparser.h"
#include "tcpmapping.h"
#include "exceptions.h"

/**
 * Method Name  MapFileParser::MapFileParser(const std::string & mapFilename)
 *
 * Parameters   const std::string & mapFilename - C++ string containing the file and path of the mappings file
 *
 * Description  Constructor
**/
MapFileParser::MapFileParser(const std::string & mapFilename) : mapFilename(mapFilename)
{
    mapFile.open(mapFilename.c_str());
    (numberOfLines = 0, skippedLines = 0, commentLines = 0, parsedLines = 0);
    mappingsList.clear();
    mapFile.exceptions(std::ifstream::badbit); // enable internal exception throwing
    parse(); // parse the provided mappings file
}

/**
 * Method Name  MapFileParser::parse()
 *
 * Description  Parses the contents of the file, storing valid mappings into mappingsList (see mapfileparser.h)
**/
void MapFileParser::parse()
{
    unsigned int scanfArgsRead = 0;

    unsigned int mapFlowCharacteristic = 0, mapStatisticsMode = 0, mapVisualisationMetric = 0;

    std::string lineToParse;

    char flowCharBuffer[51];
    char statModeBuffer[51];
    char visualParameterBuffer[51];

    // reinitialise counters
    (numberOfLines = 0, skippedLines = 0, commentLines = 0, parsedLines = 0);

    // Clear vector containing the mappings
    mappingsList.clear();

    if(mapFile.is_open())
    {
    try
    {
        while(mapFile.eof() == false)
        {
        getline(mapFile, lineToParse); // get the next line
        numberOfLines++;

        if(lineToParse[0] == '#') // skip comments
        {
            commentLines++;
            continue;
        }
        scanfArgsRead = sscanf(lineToParse.c_str(),"map %50[^:]:%50[^:]:%50s",
                       flowCharBuffer, statModeBuffer, visualParameterBuffer); // Parse it

        if(scanfArgsRead != 3) // check whether the line parsed sucessfully
        {
            skippedLines++;
            continue;
        }

        // Find the matching flow characteristics parameter for input
        for(mapFlowCharacteristic = 0; mapFlowCharacteristic < Configuration::numFlowCharacteristics; mapFlowCharacteristic++)
        {
            if(strcmp(flowCharBuffer, Configuration::flowCharacteristics[mapFlowCharacteristic]) == 0) break; // match found
        }
        if(mapFlowCharacteristic == Configuration::numFlowCharacteristics) continue; // no match found - go to the next line

        // Find matching statistics modes parameter for input
        for(mapStatisticsMode = 0; mapStatisticsMode < Configuration::numStatisticsModes; mapStatisticsMode++)
        {
            if(strcmp(statModeBuffer, Configuration::statisticsModes[mapStatisticsMode]) == 0) break; // match found
        }
        if(mapStatisticsMode == Configuration::numStatisticsModes) continue; // no match found - go to the next line

        // Find matching flow Characteristics parameter for input
        // starts at 1 to skip ("") string for l3dgecomms compatibility, see above.
        for(mapVisualisationMetric = 1; mapVisualisationMetric < Configuration::numVisualisationMetrics; mapVisualisationMetric++)
        {
            if(strcmp(visualParameterBuffer, Configuration::visualisationMetrics[mapVisualisationMetric]) == 0) break; // match found
        }
        if(mapVisualisationMetric == Configuration::numVisualisationMetrics) continue; // no match found - go to the next line

        mappingsList.push_back(TcpMapping(mapFlowCharacteristic, mapStatisticsMode, mapVisualisationMetric));
        parsedLines++;
        }
    }
    catch(std::ios_base::failure e)
    {
        std::ostringstream errorMsg;
        errorMsg << "Runtime Error : The Mappings file \"" << mapFilename << "\" is no longer accessible";
        throw fileAccessDeniedException(errorMsg.str());
    }
    }
    else
    {
    std::ostringstream errorMsg;
    errorMsg << "Initialisation failed : The Mappings file \"" << mapFilename << "\" could not be found";
    throw fileNotFoundException(errorMsg.str());
    }
}

/**
 * Method Name  std::string MapFileParser::generateReport() const
 *
 * Returns      A C++ file containing the report
 *
 * Description  Mostly for debugging - Generates a report about the parsing process
**/
std::string MapFileParser::generateReport() const
{
    std::ostringstream parseReport;

    for(unsigned int i = 0; i < mappingsList.size(); i++)
    {
    parseReport << "\tMapping the " << Configuration::statisticsModesH[mappingsList[i].statisticsMode] << " of "
            << Configuration::flowCharacteristicsH[mappingsList[i].flowCharacteristic] << " to "
            << Configuration::visualisationMetricsH[mappingsList[i].visualisationMetric] << std::endl;
    }
    parseReport << "\tParsing report: " << parsedLines << '/' << numberOfLines << " lines contained valid mappings";

    return parseReport.str();
}
