/**
 * C++ Class    Utilities
 *
 * Revision     1.0a (9th August 2010)
 * Programmer   Michael Allwright
 *
 * Purpose      Provides general functionality for the application LTCPMON. Includes statistics methods in a sub-namespace
 *              
 **/

#include <sstream>
#include <iostream>
#include <string>
#include <vector>
#include <stdexcept>
#include <algorithm>

#include <cmath>

#include "utilities.h"
#include "tcpflow.h"

/**
 * Method Name  std::string utilities::getTcpFlowID(const char localIP[], int localPort, const char remoteIP[], int remotePort)
 *
 * Parameters   const char localIP[]  - C String with local IP address
 *              int localPort         - Local Port Number
 *              const char remoteIP[] - C String with Remote IP address
 *              int remotePort        - Remote Port Number
 *
 * Returns      C++ String with the FlowID
 *
 * Description  Generates a flow ID from the 5 tuplet information
**/
std::string utilities::getTcpFlowID(const char localIP[], int localPort, const char remoteIP[], int remotePort)
{
    std::ostringstream x;
    x << "TCP " << localIP << ':' << localPort << ' ' << remoteIP << ':' << remotePort;
    return x.str();
}

/**
 * Method Name  void utilities::printFlowMapEntity(const std::pair<std::string, TcpFlow>& thePair)
 *
 * Parameters   const std::pair<std::string, TcpFlow>& thePair - C++ STL Pair containing an FlowID and the TcpFlow
 *
 * Description  Prints the FlowID alongside the number of packets in the flow to stderr
**/
void utilities::printFlowMapEntity(const std::pair<std::string, TcpFlow>& thePair)
{
    std::cerr << "Flow " << thePair.first << " contains " << thePair.second.getNumberPackets() << " packets." << std::endl;
}

/**
 * Method Name  double utilities::statistics::calcStdDev(const std::vector<unsigned int> & dataSet)
 *
 * Parameters   const std::vector<unsigned int> & dataSet - A vector of doubles i.e. that the dataset
 *
 * Description  Calculates the standard deviation
 *
 * Returns      The standard deviation
**/
double utilities::statistics::calcStdDev(const std::vector<unsigned int> & dataSet)
{
    double mean = calcMean(dataSet);
    double sumOfMeanDifferencesSquared = 0.0;

    if(dataSet.size() < 1)
    throw std::length_error("double utilities::statistics::calcStdDev(const std::vector<unsigned int> & dataSet)\n"
                "failed to calculate standard deviation for dataset with less than one element");

    for(std::vector<unsigned int>::const_iterator it = dataSet.begin(); it < dataSet.end(); it++)
      sumOfMeanDifferencesSquared += (*it - mean) * (*it - mean);

    return sqrt(sumOfMeanDifferencesSquared / dataSet.size());
}

/**
 * Method Name  double utilities::statistics::calcMin(const std::vector<unsigned int> & dataSet)
 *
 * Parameters   const std::vector<unsigned int> & dataSet - A vector of doubles i.e. that the dataset
 *
 * Description  Calculates minimum
 *
 * Returns      The minimum
**/
double utilities::statistics::calcMin(const std::vector<unsigned int> & dataSet)
{
    if(dataSet.size() < 1)
    throw std::length_error("double utilities::statistics::calcMin(const std::vector<unsigned int> & dataSet)\n"
                "failed to calculate minimum for dataset with less than one element");

    double minimum = (double)dataSet[0];

    for(std::vector<unsigned int>::const_iterator it = dataSet.begin(); it < dataSet.end(); it++)
      if(*it < minimum) minimum = *it;

    return minimum;
}

/**
 * Method Name  double utilities::statistics::calcMax(const std::vector<unsigned int> & dataSet)
 *
 * Parameters   const std::vector<unsigned int> & dataSet - A vector of doubles i.e. that the dataset
 *
 * Description  Calculates maximum
 *
 * Returns      The maximum
**/
double utilities::statistics::calcMax(const std::vector<unsigned int> & dataSet)
{
    if(dataSet.size() < 1)
    throw std::length_error("double utilities::statistics::calcMax(const std::vector<unsigned int> & dataSet)\n"
                "failed to calculate maximum for dataset with less than one element");

    double maximum = (double)dataSet[0];

    for(std::vector<unsigned int>::const_iterator it = dataSet.begin(); it < dataSet.end(); it++)
      if(*it > maximum) maximum = *it;

    return maximum;
}

/**
 * Method Name  double utilities::statistics::calcMedian(const std::vector<unsigned int> & dataSet)
 *
 * Parameters   const std::vector<unsigned int> & dataSet - A vector of doubles i.e. that the dataset
 *
 * Description  Calculates median
 *
 * Returns      The median
**/
double utilities::statistics::calcMedian(const std::vector<unsigned int> & dataSet)
{
    unsigned int dataSetSize = dataSet.size();

    if(dataSetSize < 1)
    throw std::length_error("double utilities::statistics::calcMedian(const std::vector<unsigned int> & dataSet)\n"
                "failed to calculate median for dataset with less than one element");
    std::vector<unsigned int> dataSetCopy(dataSet);

    sort(dataSetCopy.begin(),dataSetCopy.end());

    if(dataSetSize % 2 == 0)
    return (dataSetCopy[(unsigned int)(dataSetCopy.size() / 2)] + dataSetCopy[(unsigned int)((dataSetCopy.size() / 2) - 1)]) / 2;
    else
    return dataSetCopy[(unsigned int)((dataSetCopy.size() / 2) - 0.5)];
}

/**
 * Method Name  double utilities::statistics::calcMean(const std::vector<unsigned int> & dataSet)
 *
 * Parameters   const std::vector<unsigned int> & dataSet - A vector of doubles i.e. that the dataset
 *
 * Description  Calculates mean
 *
 * Returns      The mean
**/
double utilities::statistics::calcMean(const std::vector<unsigned int> & dataSet)
{
    double total = 0.0;

    if(dataSet.size() < 1)
    throw std::length_error("double utilities::statistics::calcMean(const std::vector<unsigned int> & dataSet)\n"
                "failed to calculate mean for dataset with less than one element");

    for(std::vector<unsigned int>::const_iterator it = dataSet.begin(); it < dataSet.end(); it++)
    total += *it;

    return (total / dataSet.size());
}
